/*
 * Linux Ethernet RSTP (IEEE 802.1w) bridging implementation
 * Copyright (C) 2003-2004 IMMS gGmbH, <http://www.imms.de/>
 *
 * This file implements the Port Role Transitions STM of 17.23.
 *
 * Authors:
 * Marco Goetze, <marco.goetze@imms.de>
 *
 * This code builds upon parts of the original STP bridging implementation by
 * Lennert Buytenhek, <buytenh@gnu.org>.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 */

#include <linux/kernel.h>
#include <linux/if_bridge.h>
#include <linux/smp_lock.h>
#include <asm/uaccess.h>

#include "br_private.h"
#include "br_rstp_stm.h"

#define TIMES_DIFFERENT(timer, time) \
	(((br_timer_get_residue(timer) + HZ - 1) / HZ) != (time / HZ))

static void step(struct net_bridge_port*, int*);
br_rstp_stm_port_function br_rstp_stm_port_role_transitions = &step;

static void setReRootBridge(struct net_bridge*);
static void setSyncBridge(struct net_bridge*);

static int allSynced(struct net_bridge*, port_id);
static int reRooted(struct net_bridge*, port_id);

typedef enum {
	BEGIN = BR_RSTP_STM_STATE_BEGIN,
	/* states of sub-STM for disabled, alternate, and backup roles */
	INIT_PORT,
	BLOCK_PORT,
	BACKUP_PORT,
	BLOCKED_PORT,
	/* states of sub-STM for root role */
	ROOT_PORT,
	ROOT_PROPOSED,
	ROOT_FORWARD,
	ROOT_AGREED,
	ROOT_LEARN,
	REROOT,
	REROOTED,
	/* states of sub-STM for designated role */
	DESIGNATED_PORT,
	DESIGNATED_PROPOSE,
	DESIGNATED_FORWARD,
	DESIGNATED_SYNCED,
	DESIGNATED_LEARN,
	DESIGNATED_RETIRED,
	DESIGNATED_LISTEN
} states;

static
void
enter(struct net_bridge_port *p, states *state, states target)
{
	switch (target) {

		/* states of sub-STM for disabled, alternate, and backup role */

		case INIT_PORT:
			p->role = BR_ROLE_DISABLED;
			p->synced = FALSE;
			p->sync = p->reRoot = TRUE;
			br_timer_set(&p->rrWhile, p->portTimes.FwdDelay);
			br_timer_set(&p->fdWhile, p->portTimes.FwdDelay);
			br_timer_clear(&p->rbWhile);
			break;

		case BLOCK_PORT:
			p->role = p->selectedRole;
			p->learn = p->forward = FALSE;
			break;

		case BLOCKED_PORT:
			br_timer_set(&p->fdWhile, p->portTimes.FwdDelay);
			p->synced = TRUE;
			br_timer_clear(&p->rrWhile);
			p->sync = p->reRoot = FALSE;
			break;

		case BACKUP_PORT:
			br_timer_set(&p->rbWhile, 2*p->portTimes.HelloTime);
			break;


		/* states of sub-STM for root role */

		case ROOT_PORT:
			p->role = BR_ROLE_ROOT;
			br_timer_set(&p->rrWhile, p->portTimes.FwdDelay);
			break;

		case ROOT_PROPOSED:
			setSyncBridge(p->br);
			p->proposed = FALSE;
			break;

		case ROOT_FORWARD:
			br_timer_clear(&p->fdWhile);
			p->forward = TRUE;
			break;

		case ROOT_AGREED:
			p->proposed = p->sync = FALSE;
			p->synced = TRUE;
			p->newInfo = TRUE;
			break;

		case ROOT_LEARN:
			br_timer_set(&p->fdWhile, p->portTimes.FwdDelay);
			p->learn = TRUE;
			break;

		case REROOT:
			setReRootBridge(p->br);
			break;

		case REROOTED:
			p->reRoot = FALSE;
			break;


		/* states of sub-STM for designated role */

		case DESIGNATED_PORT:
			p->role = BR_ROLE_DESIGNATED;
			break;

		case DESIGNATED_PROPOSE:
			p->proposing = TRUE;
			p->newInfo = TRUE;
			break;

		case DESIGNATED_FORWARD:
			p->forward = TRUE;
			br_timer_clear(&p->fdWhile);
			break;

		case DESIGNATED_SYNCED:
			br_timer_clear(&p->rrWhile);
			p->synced = TRUE;
			p->sync = FALSE;
			break;

		case DESIGNATED_LEARN:
			p->learn = TRUE;
			br_timer_set(&p->fdWhile, p->portTimes.FwdDelay);
			break;

		case DESIGNATED_RETIRED:
			p->reRoot = FALSE;
			break;

		case DESIGNATED_LISTEN:
			p->learn = p->forward = FALSE;
			br_timer_set(&p->fdWhile, p->portTimes.FwdDelay);
			break;

		default:
			printk(KERN_INFO "%s: bad transition target in %s: %d -> %d\n",
			       p->br->dev.name, __FILE__, *state, target);
			return;
	}
	*state = target;
}

static
void
step(struct net_bridge_port *p, int *state_val)
{
	states state = *(states*) state_val;
#define enter(target) enter(p, (states*) state_val, target)
	if (state == BEGIN ||
	    (p->role != p->selectedRole && p->selected && !p->updtInfo))
	{
		const char *s = "UNKNOWN";
		switch (p->selectedRole) {
			case BR_ROLE_DISABLED:   s = "DISABLED";   break;
			case BR_ROLE_ALTERNATE:  s = "ALTERNATE";  break;
			case BR_ROLE_BACKUP:     s = "BACKUP";     break;
			case BR_ROLE_DESIGNATED: s = "DESIGNATED"; break;
			case BR_ROLE_ROOT:       s = "ROOT";       break;
		}
		printk(KERN_INFO "%s: port %d (%s): transitioning to %s role\n",
		       p->br->dev.name, p->port_no, p->dev->name, s);
	}
	if (state != BEGIN &&
	    p->role != p->selectedRole && p->selected && !p->updtInfo)
	{
		if (p->selectedRole == BR_ROLE_ROOT)
			enter(ROOT_PORT);
		else if (p->selectedRole == BR_ROLE_DESIGNATED)
			enter(DESIGNATED_PORT);
		else
			enter(BLOCK_PORT);
	} else {
		switch (state) {
			case BEGIN:
				enter(INIT_PORT);
				break;

			/* states of sub-STM for disabled, alternate, and backup roles */

			case INIT_PORT:
				enter(BLOCK_PORT);
				break;

			case BLOCK_PORT:
				if (p->selected && !p->updtInfo && !p->learning && !p->forwarding)
					enter(BLOCKED_PORT);
				break;

			case BLOCKED_PORT:
				if (p->selected && !p->updtInfo) {
					if (TIMES_DIFFERENT(&p->fdWhile, p->portTimes.FwdDelay) ||
					    p->sync || p->reRoot || !p->synced)
					{
						enter(BLOCKED_PORT);
					} else if (TIMES_DIFFERENT(&p->rbWhile, 2*p->portTimes.HelloTime) &&
					           p->role == BR_ROLE_BACKUP)
					{
						enter(BACKUP_PORT);
					}
				}
				break;

			case BACKUP_PORT:
				enter(BLOCKED_PORT);
				break;


			/* states of sub-STM for root role */

			case ROOT_PORT:
				if (p->selected && !p->updtInfo) {
					if (p->proposed && !p->synced)
						enter(ROOT_PROPOSED);
					else if ((p->proposed || !p->synced) && allSynced(p->br, p->portId))
						enter(ROOT_AGREED);
					else if (!p->forward && !p->reRoot)
						enter(REROOT);
					else if (TIMES_DIFFERENT(&p->rrWhile, p->portTimes.FwdDelay))
						enter(ROOT_PORT);
					else if (p->reRoot && p->forward)
						enter(REROOTED);
					else if (p->br->ForceVersion >= 2 &&
					         (br_timer_has_expired(&p->fdWhile) ||
					          (br_timer_has_expired(&p->rbWhile) && reRooted(p->br, p->portId))))
					{
						if (!p->learn)
							enter(ROOT_LEARN);
						else if (p->learn && !p->forward)
							enter(ROOT_FORWARD);
					}
				}
				break;

			case ROOT_PROPOSED:
			case ROOT_FORWARD:
			case ROOT_AGREED:
			case ROOT_LEARN:
			case REROOT:
			case REROOTED:
				enter(ROOT_PORT);
				break;


			/* states of sub-STM for designated role */

			case DESIGNATED_PORT:
				if (p->selected && !p->updtInfo) {
					if (!p->forward && !p->agreed && !p->proposing && !p->operEdge)
						enter(DESIGNATED_PROPOSE);
					else if ((!p->learning && !p->forwarding && !p->synced) ||
					         (!p->synced && (p->agreed || p->operEdge)) ||
					         (p->sync && p->synced))
					{
						enter(DESIGNATED_SYNCED);
					} else if (br_timer_has_expired(&p->rrWhile) && p->reRoot)
						enter(DESIGNATED_RETIRED);
					else if (((p->sync && !p->synced) ||
					          (p->reRoot && br_timer_is_running(&p->rrWhile))) &&
					         !p->operEdge && (p->learn || p->forward))
					{
						enter(DESIGNATED_LISTEN);
					} else if ((br_timer_has_expired(&p->fdWhile) || p->agreed || p->operEdge) &&
					           (br_timer_has_expired(&p->rrWhile) || !p->reRoot) &&
					           !p->sync)
					{
						if (!p->learn)
							enter(DESIGNATED_LEARN);
						else if (p->learn && !p->forward)
							enter(DESIGNATED_FORWARD);
					}
				}
				break;

			case DESIGNATED_PROPOSE:
			case DESIGNATED_FORWARD:
			case DESIGNATED_SYNCED:
			case DESIGNATED_LEARN:
			case DESIGNATED_RETIRED:
			case DESIGNATED_LISTEN:
				enter(DESIGNATED_PORT);
				break;

			default:
				printk(KERN_INFO "%s: bad state in %s: %d; resetting to BEGIN\n",
					   p->br->dev.name, __FILE__, state);
				enter(BEGIN);
		}
	}
#undef enter
}

/* 17.19.11 */
static
void
setReRootBridge(struct net_bridge *br)
{
	struct net_bridge_port *p = br->port_list;
	while (p != NULL) {
		p->reRoot = TRUE;
		p = p->next;
	}
}

/* 17.19.10 */
static
void
setSyncBridge(struct net_bridge *br)
{
	struct net_bridge_port *p = br->port_list;
	while (p != NULL) {
		p->sync = TRUE;
		p = p->next;
	}
}

static
int
allSynced(struct net_bridge *br, port_id except)
{
	struct net_bridge_port *p = br->port_list;
	while (p != NULL) {
		if (p->portId != except && !p->synced)
			return FALSE;
		p = p->next;
	}
	return TRUE;
}

static
int
reRooted(struct net_bridge *br, port_id except)
{
	struct net_bridge_port *p = br->port_list;
	while (p != NULL) {
		if (p->portId != except && !br_timer_has_expired(&p->rrWhile))
			return FALSE;
		p = p->next;
	}
	return TRUE;
}
