/*
 * Linux Ethernet RSTP (IEEE 802.1w) bridging implementation
 * Copyright (C) 2003-2004 IMMS gGmbH, <http://www.imms.de/>
 *
 * This file implements the Topology Change STM of 17.25.
 *
 * Authors:
 * Marco Goetze, <marco.goetze@imms.de>
 *
 * This code builds upon parts of the original STP bridging implementation by
 * Lennert Buytenhek, <buytenh@gnu.org>.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 */

#include <linux/kernel.h>
#include <linux/if_bridge.h>
#include <linux/smp_lock.h>
#include <asm/uaccess.h>

#include "br_private.h"
#include "br_rstp_stm.h"

static void step(struct net_bridge_port*, int*);
br_rstp_stm_port_function br_rstp_stm_topology_change = &step;

static void flush(struct net_bridge_port*);
static int newTcWhile(struct net_bridge_port*);
static void setTcPropBridge(struct net_bridge*, port_id);

typedef enum {
	BEGIN = BR_RSTP_STM_STATE_BEGIN,
	INIT,
	INACTIVE,
	ACTIVE,
	DETECTED,
	NOTIFIED_TCN,
	NOTIFIED_TC,
	PROPAGATING,
	ACKNOWLEDGED
} states;

static
void
enter(struct net_bridge_port *p, states *state, states target)
{
	switch (target) {
		case INIT:
			flush(p);
			br_timer_clear(&p->tcWhile);
			p->tcAck = FALSE;
			break;

		case INACTIVE:
			p->rcvdTc = p->rcvdTcn = p->rcvdTcAck = p->tc = p->tcProp = FALSE;
			break;

		case ACTIVE:
			// nothing to be done
			break;

		case DETECTED:
			if (!br_timer_is_running(&p->tcWhile)) {
				printk(KERN_INFO "%s: topology change detected on port %d (%s)\n",
				       p->br->dev.name, p->port_no, p->dev->name);
			}
			br_timer_set(&p->tcWhile, newTcWhile(p));
			setTcPropBridge(p->br, p->portId);
#ifdef CONFIG_BRIDGE_CARRIER
			if (!p->carrierOK)
				flush(p);
#endif
			p->tc = FALSE;
			break;

		case NOTIFIED_TCN:
			if (!br_timer_is_running(&p->tcWhile)) {
				printk(KERN_INFO "%s: TCN received on port %d (%s)\n",
				       p->br->dev.name, p->port_no, p->dev->name);
			}
			br_timer_set(&p->tcWhile, newTcWhile(p));
			break;

		case NOTIFIED_TC:
			if (!br_timer_is_running(&p->tcWhile) && !p->tcAck) {
				printk(KERN_INFO "%s: TC received on port %d (%s)\n",
				       p->br->dev.name, p->port_no, p->dev->name);
			}
			p->rcvdTcn = p->rcvdTc = FALSE;
			if (p->role == BR_ROLE_DESIGNATED)
				p->tcAck = TRUE;
			setTcPropBridge(p->br, p->portId);
			break;

		case PROPAGATING:
			if (!br_timer_is_running(&p->tcWhile)) {
				printk(KERN_INFO "%s: propagating topology change on port %d (%s)\n",
				       p->br->dev.name, p->port_no, p->dev->name);
			}
			br_timer_set(&p->tcWhile, newTcWhile(p));
			flush(p);
			p->tcProp = FALSE;
			break;

		case ACKNOWLEDGED:
			if (br_timer_is_running(&p->tcWhile)) {
				printk(KERN_INFO "%s: topology change acknowledged on port %d (%s)\n",
				       p->br->dev.name, p->port_no, p->dev->name);
			}
			br_timer_clear(&p->tcWhile);
			p->rcvdTcAck = FALSE;
			break;

		default:
			printk(KERN_INFO "%s: bad transition target in %s: %d -> %d\n",
			       p->br->dev.name, __FILE__, *state, target);
			return;
	}
	*state = target;
}

static
void
step(struct net_bridge_port *p, int *state_val)
{
	states state = *(states*) state_val;
#define enter(target) enter(p, (states*) state_val, target)
	switch (state) {
		case BEGIN:
			enter(INIT);
			break;

		case INIT:
			enter(INACTIVE);
			break;

		case INACTIVE:
			if (p->role == BR_ROLE_ROOT || p->role == BR_ROLE_DESIGNATED)
				enter(ACTIVE);
#ifdef CONFIG_BRIDGE_CARRIER
			else if (!p->carrierOK && p->tc)
				enter(ACTIVE);
#endif
			else if (p->rcvdTc || p->rcvdTcn || p->rcvdTcAck || p->tc || p->tcProp)
				enter(INACTIVE);
			break;

		case ACTIVE:
#ifdef CONFIG_BRIDGE_CARRIER
			if (p->role != BR_ROLE_ROOT && p->role != BR_ROLE_DESIGNATED &&
			    !(!p->carrierOK && p->tc))
#else
			if (p->role != BR_ROLE_ROOT && p->role != BR_ROLE_DESIGNATED)
#endif
				enter(INIT);
			else if (p->tc)
				enter(DETECTED);
			else if (p->rcvdTcn)
				enter(NOTIFIED_TCN);
			else if (p->rcvdTc)
				enter(NOTIFIED_TC);
			else if (p->tcProp && !p->operEdge)
				enter(PROPAGATING);
			else if (p->rcvdTcAck)
				enter(ACKNOWLEDGED);
			break;

		case DETECTED:
		case NOTIFIED_TC:
		case PROPAGATING:
		case ACKNOWLEDGED:
			enter(ACTIVE);
			break;

		case NOTIFIED_TCN:
			enter(NOTIFIED_TC);
			break;

		default:
			printk(KERN_INFO "%s: bad state in %s: %d; resetting to BEGIN\n",
			       p->br->dev.name, __FILE__, state);
			enter(BEGIN);
	}
#undef enter
}

/* 17.19.6 */
static
void
flush(struct net_bridge_port *p)
{
#ifdef CONFIG_BRIDGE_CARRIER
	if (!p->operEdge || !p->carrierOK) {
#else
	if (!p->operEdge) {
#endif
		br_fdb_delete_by_port(p->br, p);
		br_fdb_insert(p->br, p, p->dev->dev_addr, TRUE);  /* add own address */
	}
}

/* 17.19.7 */
static
int
newTcWhile(struct net_bridge_port *p)
{
	return
		p->operPointToPointMAC && p->sendRSTP
			? 2 * p->portTimes.HelloTime
			: p->br->rootTimes.MaxAge + p->br->rootTimes.FwdDelay;
}

/* 17.19.14 */
static
void
setTcPropBridge(struct net_bridge *br, port_id except)
{
	struct net_bridge_port *p = br->port_list;
	while (p != NULL) {
		if (p->portId != except)
			p->tcProp = TRUE;
		p = p->next;
	}
}
